import json as _json

from datetime import datetime, date
from typing   import Any

def _default(o: Any) -> str:
    return {
        datetime: lambda o: o.strftime('%Y-%m-%d %H:%M:%S'),
        date    : lambda o: o.strftime('%Y-%m-%d')
    }.get(type(o), str)(o)


def dumps(*args: Any, **kwargs: Any) -> str:
    if not 'default' in kwargs:
        kwargs['default'] = _default
        pass
    return _json.dumps(*args, **kwargs)



from typing import Literal

from pydantic import BaseModel, StrictStr

from nsj_flask_utils.dtos.base_dto import BaseDTO


class PaginationDTO(BaseModel):
    order: Literal['ASC', 'DESC'] = 'DESC'
    order_by: StrictStr = ''
    page: int = 1
    limit: int = 20


class PaginationParams:
    data: PaginationDTO
    offset: int

    def __init__(self, dto_cls: type[BaseDTO]
                     , data: PaginationDTO) -> None:
        self.data = data
        self.offset = data.limit * (data.page - 1)

        if not data.order_by:
            self.order_by = dto_cls.__order_by__
        elif data.order_by in dto_cls.model_fields.keys():
            self.order_by = data.order_by
        else:
            self.order_by = dto_cls.__order_by__
            pass
        pass

    def to_dict(self) -> dict[str, str | int]:
        return {
            "limit": self.data.limit,
            "offset": self.offset,
            "order_by": self.order_by,
            "order": self.data.order,
            "page": self.data.page,
        }
    pass



from typing import ClassVar

from pydantic import BaseModel


class BaseDTO(BaseModel):
    __order_by__  : ClassVar[str]
    __table_name__: ClassVar[str]
    __schema__    : ClassVar[str]
    __PK__        : ClassVar[str]
    pass



from typing import Any

import nsj_flask_utils.utils.json as json


class HTTPBaseException(Exception):
    data: dict[str, Any]
    code: int

    def __init__(self,data: dict[str, Any]) -> None:
        self.data = data

    @classmethod
    def message(cls: type['HTTPBaseException'],
                message: str) -> 'HTTPBaseException':
        return cls({'message': message})

    def to_tuple(self) -> tuple[dict[str, Any], int]:
        return (self.data, self.code)

    def to_tuple_str_int(self) -> tuple[str, int]:
        return (json.dumps(self.data), self.code)


class HTTP400(HTTPBaseException):
    code = 400



from .request_args      import RequestArgs
from .pagination_params import PaginationParams
from .filter_params     import FilterParams

__all__ = [
    'RequestArgs',
    'PaginationParams',
    'FilterParams'
]



import math
from typing import Any, Generic, List, TypeVar

from sqlalchemy.engine   import Connection
from werkzeug.exceptions import NotFound

from nsj_flask_utils.utils import RequestArgs
from nsj_flask_utils.daos  import BaseDAO
from nsj_flask_utils.dtos  import BaseDTO

TDTO = TypeVar('TDTO', bound=BaseDTO)
TDAO = TypeVar('TDAO', bound=BaseDAO)


class BaseService(Generic[TDAO, TDTO]):
    _dao_cls: type[TDAO]
    _dto_cls: type[TDTO]

    _dao: TDAO
    _dto: type[TDTO]

    def __init__(self, db : Connection
                     , dao: type[TDAO] | None = None
                     , dto: type[TDTO] | None = None) -> None:
        if dao:
            self._dao = dao(db)
        elif self._dao_cls:
            self._dao = self._dao_cls(db)
        else:
            raise Exception('DAO not defined')

        if dto:
            self._dto = dto
        elif self._dto_cls:
            self._dto = self._dto_cls
        else:
            raise Exception('DTO not defined')
        pass

    def list(self, req_args: RequestArgs) -> list[TDTO]:
        return self._dao.list(req_args)

    def create(self, dto: TDTO) -> TDTO:
        return self._dao.create(dto)

    def get_or_none(self, id: Any) -> TDTO | None:
        return self._dao.get(id)

    def get_or_raise(self, id: Any
                         , exception: type[Exception] = NotFound) -> TDTO:
        rs = self._dao.get(id)
        if not rs:
            raise exception("Not found")
        return rs

    def update(self, id: Any, dto: TDTO) -> TDTO:
        original = self.get_or_raise(id)
        atualizado = original.model_copy(
            update=dto.model_dump(exclude_unset=True)
        )
        self._dao.update(atualizado)
        return atualizado

    def delete(self, id: Any) -> TDTO:
        self.get_or_raise(id)
        return self._dao.delete(id)

    def paginate_response(self, req_args: RequestArgs
                              , data    : List[TDTO]
                              , exclude : set[str] | None = None
                              , include : set[str] | None = None) -> dict[str, Any]:
        rows_total = self._dao.get_rows_total(req_args)
        pagination_params = req_args.get_pagination_params()
        current_page = int(pagination_params["page"])
        total_pages = math.ceil(rows_total / int(pagination_params["limit"]))

        next_page: int | None = None
        if current_page < total_pages:
            next_page = current_page + 1
            pass

        prev_page: int | None = None
        if current_page != 1:
            prev_page = current_page - 1
            pass

        if include:
            ndata = [model.model_dump(include=include) for model in data]
        elif exclude:
            ndata = [model.model_dump(exclude=exclude) for model in data]
        else:
            ndata = [model.model_dump() for model in data]

        response = {
            "total": rows_total,
            "next_page": next_page,
            "current_page": current_page,
            "total_pages": total_pages,
            "prev_page": prev_page,
            "data": ndata,
        }

        return response



from typing import Any

from flask import Flask, Response
from werkzeug.exceptions import HTTPException

from nsj_flask_utils.utils      import json
from nsj_flask_utils.exceptions import HTTPBaseException


def register_error_handlers(app: Flask) -> None:
    @app.errorhandler(HTTPException)
    def http_exception_handler(e: HTTPException) -> Response:
        """Return JSON instead of HTML for HTTP errors."""
        eres = e.get_response()
        response = Response(None, eres.status_code, eres.headers
                                , eres.mimetype, eres.content_type)
        response.data = json.dumps(
            {
                "code": e.code,
                "name": e.name,
                "description": e.description,
            }
        )
        response.content_type = "application/json"
        return response

    @app.errorhandler(HTTPException)
    def http_base_exception_handler(e: HTTPBaseException) -> tuple[str, int]:
        """Return JSON instead of HTML for HTTP errors."""
        return e.to_tuple_str_int()
    pass



from typing import Generic, Any, Mapping, Generic, TypeVar

import sqlalchemy
from werkzeug.exceptions import NotFound
from sqlalchemy.engine   import Connection, CursorResult, Transaction

from nsj_flask_utils.dtos.base_dto      import BaseDTO
from nsj_flask_utils.utils.request_args import RequestArgs

TDTO = TypeVar('TDTO', bound=BaseDTO)


class BaseDAO(Generic[TDTO]):
    _db : Connection
    _dto: type[TDTO]

    def __init__(self, db : Connection
                     , dto: type[TDTO] | None = None) -> None:
        self._db = db
        if dto:
            self._dto = dto
            pass
        pass

    def _execute(self, query: str
                     , params: Mapping[str, Any] = {}) -> CursorResult[Any]:
        tran: Transaction = self._db.begin()
        rs = self._db.execute(sqlalchemy.text(query), params)
        tran.commit()
        return rs

    def _get_ftbl_name(self) -> str:
        """Returns the table name formatede: `schema.tbl_name'"""
        return f"{self._dto.__schema__}.{self._dto.__table_name__}"

    def _list(self, query: str
                  , pagi_args: dict[str, str | int]
                  , filter_args: dict[str, str]
                  , group_by: str | None = None
                  , special_map: dict[str, str] = {}) -> list[TDTO]:
        query = self._filter_query(query, filter_args, group_by, special_map)

        if group_by:
            query += f"\nGROUP BY {group_by}"
            pagi_args['order_by'] = group_by
            pass

        query += f"""
ORDER BY {pagi_args['order'],} {pagi_args['order_by'],}
LIMIT :limit
OFFSET :offset
"""

        params = {
            **pagi_args,
            **filter_args,
        }

        rs = self._execute(query, params)
        return [
            self._dto(**x)
            for x in rs.mappings().all()
        ]

    def get(self, id: Any) -> TDTO | None:
        query: str = f"SELECT * FROM {self._get_ftbl_name()}"
        filter_args = {self._dto.__PK__: id}
        query = self._filter_query(query, filter_args)
        rs = self._execute(query, filter_args).mappings().first()
        if not rs:
            return None
        return self._dto(**rs)

    def _filter_query(self, query: str
                          , filter_args: dict[str, str]
                          , group_by: str | None = None
                          , special_map: dict[str, str] = {}) -> str:
        def sp_get(k: str) -> str:
            return special_map.get(k, k)

        keys = list(filter_args.keys())

        if group_by:
            query += f"\nWHERE {group_by} IS NOT NULL"
            pass

        if keys:
            if not group_by:
                query += f"\nWHERE {sp_get(keys[0])} = :{keys[0]}"
                keys.pop(0)
                pass
            for key in keys:
                query += f"\n  AND {sp_get(key)} = :{key}"
                pass
            pass
        return query

    def _get_first(self, query: str
                       , filter_args: dict[str, str]) -> Any | None:
        rs = self._execute(query, filter_args)
        frow = rs.first()
        if not frow:
            return None
        return frow[0]

    def get_all(self) -> list[TDTO]:
        query: str = f"SELECT * FROM {self._get_ftbl_name()}"
        rs = self._execute(query)
        return [
            self._dto(**x)
            for x in rs.mappings().all()
        ]

    def list(self, req_args: RequestArgs) -> list[TDTO]:
        query: str = f"SELECT * FROM {self._get_ftbl_name()}"
        return self._list(query, req_args.get_pagination_params()
                               , req_args.get_filter_params())

    def get_rows_total(self, req_args: RequestArgs
                           , special_map: dict[str, str] = {}) -> int:
        filter_args = req_args.get_filter_params()

        query: str = f"SELECT COUNT(*) FROM {self._get_ftbl_name()}"

        gb: str | None = req_args.group_by
        if gb:
            query = f"SELECT COUNT(DISTINCT {gb}) FROM {self._get_ftbl_name()}"
            pass
        query = self._filter_query(query, filter_args, gb, special_map)

        ret = self._get_first(query, filter_args)
        if not ret:
            return -1
        return int(ret)
